<?php

namespace Drupal\votingapi_reaction;

use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Render\Renderer;
use Drupal\Core\Session\AccountProxy;
use Drupal\votingapi\Entity\Vote;
use Drupal\votingapi\Entity\VoteType;
use Drupal\votingapi\VoteResultFunctionManager;
use Drupal\votingapi_reaction\Plugin\Field\FieldType\VotingApiReactionItemInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\file\Entity\File;

/**
 * Manages reactions through Voting API entities.
 */
class VotingApiReactionManager implements ContainerInjectionInterface {

  /**
   * @var \Drupal\votingapi\VoteStorage $voteStorage
   */
  protected $voteStorage;

  /**
   * @var \Drupal\Core\Entity\Sql\SqlContentEntityStorage $voteTypeStorage
   */
  protected $voteTypeStorage;

  /**
   * @var \Drupal\Core\Session\AccountProxy $currentUser
   */
  protected $currentUser;

  /**
   * @var \Drupal\votingapi\VoteResultFunctionManager $votingApiResults
   */
  protected $votingApiResults;

  /**
   * @var \Drupal\Core\Render\Renderer $renderer
   */
  protected $renderer;

  /**
   * Class constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   * @param \Drupal\Core\Session\AccountProxy $currentUser
   * @param \Drupal\votingapi\VoteResultFunctionManager $votingApiResults
   * @param \Drupal\Core\Render\Renderer $renderer
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function __construct(EntityTypeManagerInterface $entityTypeManager, AccountProxy $currentUser, VoteResultFunctionManager $votingApiResults, Renderer $renderer) {
    $this->voteStorage = $entityTypeManager->getStorage('vote');
    $this->voteTypeStorage = $entityTypeManager->getStorage('vote_type');
    $this->currentUser = $currentUser;
    $this->votingApiResults = $votingApiResults;
    $this->renderer = $renderer;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('current_user'),
      $container->get('plugin.manager.votingapi.resultfunction'),
      $container->get('renderer')
    );
  }

  /**
   * Load previous reaction of the user for certain field.
   *
   * @param \Drupal\votingapi\Entity\Vote $entity
   * @param $settings
   *
   * @return \Drupal\Core\Entity\EntityInterface|null
   */
  public function lastReaction(Vote $entity, $settings) {
    $query = $this->voteStorage->getQuery()
      ->condition('entity_id', $entity->getVotedEntityId())
      ->condition('entity_type', $entity->getVotedEntityType())
      ->condition('field_name', $entity->get('field_name')->value)
      ->condition('user_id', $this->currentUser->id());

    if ($this->currentUser->isAnonymous()) {
      // Filter by Cookie method
      if (in_array(VotingApiReactionItemInterface::BY_COOKIES, $settings['anonymous_detection'])) {
        $query->condition('id', $this->recallReaction($entity));
      }
      // Filter by IP method
      if (in_array(VotingApiReactionItemInterface::BY_IP, $settings['anonymous_detection'])) {
        $query->condition('vote_source', Vote::getCurrentIp());
      }

      // Filter by rollover
      $rollover = $settings['anonymous_rollover'];
      if ($rollover == VotingApiReactionItemInterface::VOTINGAPI_ROLLOVER) {
        $rollover = \Drupal::configFactory()
          ->get('votingapi.settings')
          ->get('anonymous_window');
      }
      if ($rollover != VotingApiReactionItemInterface::NEVER_ROLLOVER) {
        $query->condition('timestamp', time() - $rollover, '>=');
      }
    }

    $ids = $query->execute();

    return $this->voteStorage->load(intval(array_pop($ids)));
  }

  /**
   * Return voting results for each active reaction.
   *
   * @param \Drupal\votingapi\Entity\Vote $entity
   * @param $settings
   *
   * @return array
   */
  public function getResults(Vote $entity, $settings) {
    // Get results for each reaction
    $results = $this->votingApiResults->getResults($entity->getVotedEntityType(), $entity->getVotedEntityId());
    $reactions = array_filter($settings['reactions']);

    return array_intersect_key($results, $reactions);
  }

  /**
   * Recalculate results for given reaction and entity.
   *
   * @param $entity_type
   * @param $entity_id
   * @param $type
   */
  public function recalculateResults($entity_type, $entity_id, $type) {
    $this->votingApiResults->recalculateResults($entity_type, $entity_id, $type);
  }

  /**
   * Return all vote types marked as reaction.
   *
   * @return array
   */
  public function allReactions() {
    return array_filter($this->voteTypeStorage->loadMultiple(), function (VoteType $entity) {
      return $entity->getThirdPartySetting('votingapi_reaction', 'reaction');
    });
  }

  /**
   * Return rendered list of active reactions.
   *
   * @param $settings
   * @param $results
   *
   * @return array
   */
  public function getReactions($settings, $results) {
    // Get only enabled reactions
    $entities = array_filter($this->allReactions(), function (VoteType $entity) use ($settings) {
      return in_array($entity->id(), array_filter($settings['reactions']));
    });

    // Configure the object
    $reactions = array_map(function (VoteType $entity) use ($settings, $results) {
      $reaction = [
        '#theme' => 'votingapi_reaction_item',
        '#reaction' => $entity->id(),
      ];

      if ($settings['show_icon']) {
        $reaction['#icon'] = $this->getIcon($entity);
      }

      if ($settings['show_label']) {
        $reaction['#label'] = $entity->label();
      }

      if ($settings['show_count']) {
        $reaction['#count'] = isset($results[$entity->id()]['vote_sum'])
          ? $results[$entity->id()]['vote_sum']
          : 0;
      }

      return $reaction;
    }, $entities);

    // Sorting
    if ($settings['sort_reactions'] != 'none') {
      uasort($reactions, function ($a, $b) use ($settings, $results) {
        $count_a = isset($results[$a['#reaction']]['vote_count'])
          ? $results[$a['#reaction']]['vote_count'] : 0;
        $count_b = isset($results[$b['#reaction']]['vote_count'])
          ? $results[$b['#reaction']]['vote_count'] : 0;

        if ($settings['sort_reactions'] == 'desc') {
          return $count_a > $count_b ? -1 : 1;
        }
        else {
          return $count_a < $count_b ? -1 : 1;
        }
      });
    }

    // Render the
    return array_map(function ($reaction) {
      return $this->renderer->render($reaction);
    }, $reactions);
  }

  /**
   * Store reaction from session variable.
   *
   * @param \Drupal\votingapi\Entity\Vote $entity
   */
  public function rememberReaction(Vote $entity) {
    $_SESSION['votingapi_reaction'][implode(':', [
      $entity->getVotedEntityId(),
      $entity->getVotedEntityType(),
      $entity->get('field_name')->value,
    ])] = $entity->id();
  }

  /**
   * Remove reaction from session variable.
   *
   * @param \Drupal\votingapi\Entity\Vote $entity
   */
  public function forgetReaction(Vote $entity) {
    unset($_SESSION['votingapi_reaction'][implode(':', [
        $entity->getVotedEntityId(),
        $entity->getVotedEntityType(),
        $entity->get('field_name')->value,
      ])]);
  }

  /**
   * Restore reaction from session variable.
   *
   * @param \Drupal\votingapi\Entity\Vote $entity
   *
   * @return string|null
   */
  public function recallReaction(Vote $entity) {
    $key = implode(':', [
      $entity->getVotedEntityId(),
      $entity->getVotedEntityType(),
      $entity->get('field_name')->value,
    ]);
    return !empty($_SESSION['votingapi_reaction'][$key])
      ? $_SESSION['votingapi_reaction'][$key]
      : NULL;
  }

  /**
   * Return URL to reaction icon.
   *
   * @param \Drupal\votingapi\Entity\VoteType $entity
   *
   * @return string
   */
  public function getIcon(VoteType $entity, &$default = TRUE) {
    $path = implode('/', [
      '', // $GLOBAL['base_path']
      drupal_get_path('module', 'votingapi_reaction'),
      'svg',
      '', // Trailing slash
    ]);

    // Fallback icon
    $url = $path . 'reaction_noicon.svg';
    // User defined icon
    $icon = $entity->getThirdPartySetting('votingapi_reaction', 'icon');
    if ($icon && $file = File::load($icon)) {
      $url = file_create_url($file->getFileUri());
      $default = FALSE;
    }
    // Default icon
    elseif (file_exists(DRUPAL_ROOT . $path . $entity->id() . '.svg')) {
      $url = $path . $entity->id() . '.svg';
    }

    return $url;
  }

}
